import base64
import uuid
from collections.abc import Callable
from pathlib import Path
from typing import cast

from openai import AsyncOpenAI
from openai.types.responses import Response
from openai.types.responses.tool_param import CodeInterpreter, ToolParam

from ragbits.agents.tools.types import ImageGenerationResponse
from ragbits.core.sources.base import get_local_storage_dir


def get_web_search_tool(model_name: str, api_key: str | None = None, additional_params: dict | None = None) -> Callable:
    """
    Returns a native OpenAI web search tool as function

    Args:
        model_name: The name of the model
        api_key: The API key for the model
        additional_params: The additional tool parameters

    Returns:
        web search function
    """
    params_to_pass = additional_params if additional_params else {}
    tool_object = OpenAITools(model_name, api_key, {"type": "web_search_preview", **params_to_pass})
    return tool_object.search_web


def get_code_interpreter_tool(
    model_name: str, api_key: str | None = None, additional_params: dict | None = None
) -> Callable:
    """
    Returns a native OpenAI code interpreter tool as function

    Args:
        model_name: The name of the model
        api_key: The API key for the model
        additional_params: The additional tool parameters

    Returns:
        code interpreter function
    """
    params_to_pass = additional_params if additional_params else {}
    tool_object = OpenAITools(
        model_name, api_key, cast(CodeInterpreter, {"type": "code_interpreter", **params_to_pass})
    )
    return tool_object.code_interpreter


def get_image_generation_tool(
    model_name: str, api_key: str | None = None, additional_params: dict | None = None
) -> Callable:
    """
    Returns a native OpenAI image generation tool as function

    Args:
        model_name: The name of the model
        api_key: The API key for the model
        additional_params: The additional tool parameters

    Returns:
        image generation function
    """
    params_to_pass = additional_params if additional_params else {}
    tool_object = OpenAITools(model_name, api_key, {"type": "image_generation", **params_to_pass})
    return tool_object.image_generation


class OpenAIResponsesLLM:
    """
    Class serving as a wrapper for tool calls to responses API of OpenAI
    """

    def __init__(self, model_name: str, api_key: str | None, tool_param: ToolParam):
        self._client = AsyncOpenAI(api_key=api_key)
        self._model_name = model_name
        self._tool_param = tool_param

    async def use_tool(self, query: str) -> Response:
        """
        Uses tool based on query and returns output.

        Args:
            query: query for the tool

        Returns:
            Output of the tool
        """
        return await self._client.responses.create(
            model=self._model_name,
            tools=[self._tool_param],
            tool_choice="required",
            input=query,
        )


class OpenAITools:
    """
    Class wrapping tool calls to responses API of OpenAI
    """

    AVAILABLE_TOOLS = {"web_search_preview", "code_interpreter", "image_generation"}

    def __init__(self, model_name: str, api_key: str | None, tool_param: ToolParam):
        self._responses_llm = OpenAIResponsesLLM(model_name, api_key, tool_param)

    async def search_web(self, query: str) -> Response:
        """
        Searches web for a query

        Args:
            query: The query to search

        Returns:
            The web search result
        """
        return await self._responses_llm.use_tool(query)

    async def code_interpreter(self, query: str) -> Response:
        """
        Performs actions in code interpreter based on query

        Args:
            query: The query with instructions

        Returns:
            Output of the interpreter
        """
        return await self._responses_llm.use_tool(query)

    async def image_generation(self, query: str, save_path: str | Path | None = None) -> ImageGenerationResponse:
        """
        Generate image based on query.

        Args:
            query: The query with instructions
            save_path: The path to save the generated image

        Returns:
            LLM dict with image_path and output_text
        """
        if save_path is None:
            save_path = get_local_storage_dir() / f"{uuid.uuid4()}.png"

        response = await self._responses_llm.use_tool(query)
        image_data = next((output.result for output in response.output if output.type == "image_generation_call"), None)

        if image_data:
            with open(save_path, "wb") as f:
                f.write(base64.b64decode(image_data))
            return ImageGenerationResponse(image_path=save_path, output_text=response.output_text)
        else:
            return ImageGenerationResponse(image_path=None, output_text=response.output_text)



import asyncio
import types
import uuid
from collections.abc import AsyncGenerator, AsyncIterator, Callable
from contextlib import suppress
from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from inspect import iscoroutinefunction
from types import ModuleType, SimpleNamespace
from typing import Any, ClassVar, Generic, TypeVar, cast, overload

from pydantic import (
    BaseModel,
    Field,
)
from typing_extensions import Self

from ragbits import agents
from ragbits.agents.exceptions import (
    AgentInvalidPromptInputError,
    AgentMaxTokensExceededError,
    AgentMaxTurnsExceededError,
    AgentNextPromptOverLimitError,
    AgentToolDuplicateError,
    AgentToolExecutionError,
    AgentToolNotAvailableError,
    AgentToolNotSupportedError,
)
from ragbits.agents.mcp.server import MCPServer, MCPServerStdio, MCPServerStreamableHttp
from ragbits.agents.mcp.utils import get_tools
from ragbits.agents.tool import Tool, ToolCallResult, ToolChoice
from ragbits.core.audit.traces import trace
from ragbits.core.llms.base import LLM, LLMClientOptionsT, LLMOptions, LLMResponseWithMetadata, ToolCall, Usage
from ragbits.core.options import Options
from ragbits.core.prompt.base import BasePrompt, ChatFormat, SimplePrompt
from ragbits.core.prompt.prompt import Prompt, PromptInputT, PromptOutputT
from ragbits.core.types import NOT_GIVEN, NotGiven
from ragbits.core.utils.config_handling import ConfigurableComponent
from ragbits.core.utils.decorators import requires_dependencies

with suppress(ImportError):
    from a2a.types import AgentCapabilities, AgentCard, AgentSkill
    from pydantic_ai import Agent as PydanticAIAgent
    from pydantic_ai import mcp

    from ragbits.core.llms import LiteLLM

_Input = TypeVar("_Input", bound=BaseModel)
_Output = TypeVar("_Output")


@dataclass
class AgentResult(Generic[PromptOutputT]):
    """
    Result of the agent run.
    """

    content: PromptOutputT
    """The output content of the agent."""
    metadata: dict
    """The additional data returned by the agent."""
    history: ChatFormat
    """The history of the agent."""
    tool_calls: list[ToolCallResult] | None = None
    """Tool calls run by the agent."""
    usage: Usage = Field(default_factory=Usage)
    """The token usage of the agent run."""


class AgentOptions(Options, Generic[LLMClientOptionsT]):
    """
    Options for the agent run.
    """

    llm_options: LLMClientOptionsT | None | NotGiven = NOT_GIVEN
    """The options for the LLM."""
    max_turns: int | None | NotGiven = NOT_GIVEN
    """The maximum number of turns the agent can take, if NOT_GIVEN,
    it defaults to 10, if None, agent will run forever"""
    max_total_tokens: int | None | NotGiven = NOT_GIVEN
    """The maximum number of tokens the agent can use, if NOT_GIVEN
    or None, agent will run forever"""
    max_prompt_tokens: int | None | NotGiven = NOT_GIVEN
    """The maximum number of prompt tokens the agent can use, if NOT_GIVEN
    or None, agent will run forever"""
    max_completion_tokens: int | None | NotGiven = NOT_GIVEN
    """The maximum number of completion tokens the agent can use, if NOT_GIVEN
    or None, agent will run forever"""


DepsT = TypeVar("DepsT")


class AgentDependencies(BaseModel, Generic[DepsT]):
    """
    Container for agent runtime dependencies.

    Becomes immutable after first attribute access.
    """

    model_config = {"arbitrary_types_allowed": True}

    _frozen: bool
    _value: DepsT | None

    def __init__(self, value: DepsT | None = None) -> None:
        super().__init__()
        self._value = value
        self._frozen = False

    def __setattr__(self, name: str, value: object) -> None:
        is_frozen = False
        if name != "_frozen":
            try:
                is_frozen = object.__getattribute__(self, "_frozen")
            except AttributeError:
                is_frozen = False

        if is_frozen and name not in {"_frozen"}:
            raise RuntimeError("Dependencies are immutable after first access")

        super().__setattr__(name, value)

    @property
    def value(self) -> DepsT | None:
        return self._value

    @value.setter
    def value(self, value: DepsT) -> None:
        if self._frozen:
            raise RuntimeError("Dependencies are immutable after first access")
        self._value = value

    def _freeze(self) -> None:
        if not self._frozen:
            self._frozen = True

    def __getattr__(self, name: str) -> object:
        value = object.__getattribute__(self, "_value")
        if value is None:
            raise AttributeError(name)
        self._freeze()
        return getattr(value, name)

    def __contains__(self, key: str) -> bool:
        value = object.__getattribute__(self, "_value")
        return hasattr(value, key) if value is not None else False


class AgentRunContext(BaseModel, Generic[DepsT]):
    """Context for the agent run."""

    deps: AgentDependencies[DepsT] = Field(default_factory=lambda: AgentDependencies())
    """Container for external dependencies."""
    usage: Usage = Field(default_factory=Usage)
    """The usage of the agent."""


class AgentResultStreaming(AsyncIterator[str | ToolCall | ToolCallResult]):
    """
    An async iterator that will collect all yielded items by LLM.generate_streaming(). This object is returned
    by `run_streaming`. It can be used in an `async for` loop to process items as they arrive. After the loop completes,
    all items are available under the same names as in AgentResult class.
    """

    def __init__(
        self, generator: AsyncGenerator[str | ToolCall | ToolCallResult | SimpleNamespace | BasePrompt | Usage]
    ):
        self._generator = generator
        self.content: str = ""
        self.tool_calls: list[ToolCallResult] | None = None
        self.metadata: dict = {}
        self.history: ChatFormat
        self.usage: Usage = Usage()

    def __aiter__(self) -> AsyncIterator[str | ToolCall | ToolCallResult]:
        return self

    async def __anext__(self) -> str | ToolCall | ToolCallResult:
        try:
            item = await self._generator.__anext__()
            match item:
                case str():
                    self.content += item
                case ToolCall():
                    pass
                case ToolCallResult():
                    if self.tool_calls is None:
                        self.tool_calls = []
                    self.tool_calls.append(item)
                case BasePrompt():
                    item.add_assistant_message(self.content)
                    self.history = item.chat
                    item = await self._generator.__anext__()
                    item = cast(SimpleNamespace, item)
                    item.result = {
                        "content": self.content,
                        "metadata": self.metadata,
                        "tool_calls": self.tool_calls,
                    }
                    raise StopAsyncIteration
                case Usage():
                    self.usage = item
                    return await self.__anext__()
                case _:
                    raise ValueError(f"Unexpected item: {item}")
            return item
        except StopAsyncIteration:
            raise


class Agent(
    ConfigurableComponent[AgentOptions[LLMClientOptionsT]], Generic[LLMClientOptionsT, PromptInputT, PromptOutputT]
):
    """
    Agent class that orchestrates the LLM and the prompt, and can call tools.

    Current implementation is highly experimental, and the API is subject to change.
    """

    options_cls: type[AgentOptions] = AgentOptions
    default_module: ClassVar[ModuleType | None] = agents
    configuration_key: ClassVar[str] = "agent"
    # Syntax sugar fields
    user_prompt: ClassVar[str] = "{{ input }}"
    system_prompt: ClassVar[str | None] = None
    input_type: PromptInputT | None = None
    prompt_cls: ClassVar[type[Prompt] | None] = None

    def __init__(
        self,
        llm: LLM[LLMClientOptionsT],
        name: str | None = None,
        description: str | None = None,
        prompt: str | type[Prompt[PromptInputT, PromptOutputT]] | Prompt[PromptInputT, PromptOutputT] | None = None,
        *,
        history: ChatFormat | None = None,
        keep_history: bool = False,
        tools: list[Callable] | None = None,
        mcp_servers: list[MCPServer] | None = None,
        default_options: AgentOptions[LLMClientOptionsT] | None = None,
    ) -> None:
        """
        Initialize the agent instance.

        Args:
            llm: The LLM to run the agent.
            name: Optional name of the agent. Used to identify the agent instance.
            description: Optional description of the agent.
            prompt: The prompt for the agent. Can be:
                - str: A string prompt that will be used as system message when combined with string input,
                    or as the user message when no input is provided during run().
                - type[Prompt]: A structured prompt class that will be instantiated with the input.
                - Prompt: Already instantiated prompt instance
                - None: No predefined prompt. The input provided to run() will be used as the complete prompt.
            history: The history of the agent.
            keep_history: Whether to keep the history of the agent.
            tools: The tools available to the agent.
            mcp_servers: The MCP servers available to the agent.
            default_options: The default options for the agent run.
        """
        super().__init__(default_options)
        self.id = uuid.uuid4().hex[:8]
        self.llm = llm
        self.prompt = prompt
        self.name = name
        self.description = description
        self.tools = []
        for tool in tools or []:
            if isinstance(tool, tuple):
                agent, kwargs = tool
                self.tools.append(Tool.from_agent(agent, **kwargs))
            elif isinstance(tool, Agent):
                self.tools.append(Tool.from_agent(tool))
            else:
                self.tools.append(Tool.from_callable(tool))
        self.mcp_servers = mcp_servers or []
        self.history = history or []
        self.keep_history = keep_history

        if getattr(self, "system_prompt", None) and not getattr(self, "input_type", None):
            raise ValueError(
                f"Agent {type(self).__name__} defines a system_prompt but has no input_type. ",
                "Use Agent.prompt decorator to properly assign it.",
            )

    @overload
    async def run(
        self: "Agent[LLMClientOptionsT, None, PromptOutputT]",
        input: str | None = None,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResult[PromptOutputT]: ...

    @overload
    async def run(
        self: "Agent[LLMClientOptionsT, PromptInputT, PromptOutputT]",
        input: PromptInputT,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResult[PromptOutputT]: ...

    async def run(
        self,
        input: str | PromptInputT | None = None,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResult[PromptOutputT]:
        """
        Run the agent. The method is experimental, inputs and outputs may change in the future.

        Args:
            input: The input for the agent run. Can be:
                - str: A string input that will be used as user message.
                - PromptInputT: Structured input for use with structured prompt classes.
                - None: No input. Only valid when a string prompt was provided during initialization.
            options: The options for the agent run.
            context: The context for the agent run.
            tool_choice: Parameter that allows to control what tool is used at first call. Can be one of:
                - "auto": let model decide if tool call is needed
                - "none": do not call tool
                - "required: enforce tool usage (model decides which one)
                - Callable: one of provided tools

        Returns:
            The result of the agent run.

        Raises:
            AgentToolDuplicateError: If the tool names are duplicated.
            AgentToolNotSupportedError: If the selected tool type is not supported.
            AgentToolNotAvailableError: If the selected tool is not available.
            AgentInvalidPromptInputError: If the prompt/input combination is invalid.
            AgentMaxTurnsExceededError: If the maximum number of turns is exceeded.
        """
        if context is None:
            context = AgentRunContext()

        input = cast(PromptInputT, input)
        merged_options = (self.default_options | options) if options else self.default_options
        llm_options = merged_options.llm_options or self.llm.default_options

        prompt_with_history = self._get_prompt_with_history(input)
        tools_mapping = await self._get_all_tools()
        tool_calls = []

        turn_count = 0
        max_turns = merged_options.max_turns
        max_turns = 10 if max_turns is NOT_GIVEN else max_turns
        with trace(input=input, options=merged_options) as outputs:
            while not max_turns or turn_count < max_turns:
                self._check_token_limits(merged_options, context.usage, prompt_with_history, self.llm)
                response = cast(
                    LLMResponseWithMetadata[PromptOutputT],
                    await self.llm.generate_with_metadata(
                        prompt=prompt_with_history,
                        tools=[tool.to_function_schema() for tool in tools_mapping.values()],
                        tool_choice=tool_choice if tool_choice and turn_count == 0 else None,
                        options=self._get_llm_options(llm_options, merged_options, context.usage),
                    ),
                )
                context.usage += response.usage or Usage()

                if not response.tool_calls:
                    break

                for tool_call in response.tool_calls:
                    result = await self._execute_tool(tool_call=tool_call, tools_mapping=tools_mapping, context=context)
                    tool_calls.append(result)

                    prompt_with_history = prompt_with_history.add_tool_use_message(**result.__dict__)

                turn_count += 1
            else:
                raise AgentMaxTurnsExceededError(cast(int, max_turns))

            outputs.result = {
                "content": response.content,
                "metadata": response.metadata,
                "tool_calls": tool_calls or None,
            }

            prompt_with_history = prompt_with_history.add_assistant_message(response.content)

            if self.keep_history:
                self.history = prompt_with_history.chat

            return AgentResult(
                content=response.content,
                metadata=response.metadata,
                tool_calls=tool_calls or None,
                history=prompt_with_history.chat,
                usage=context.usage,
            )

    @overload
    def run_streaming(
        self: "Agent[LLMClientOptionsT, None, PromptOutputT]",
        input: str | None = None,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResultStreaming: ...

    @overload
    def run_streaming(
        self: "Agent[LLMClientOptionsT, PromptInputT, PromptOutputT]",
        input: PromptInputT,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResultStreaming: ...

    def run_streaming(
        self,
        input: str | PromptInputT | None = None,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AgentResultStreaming:
        """
        This method returns an `AgentResultStreaming` object that can be asynchronously
        iterated over. After the loop completes, all items are available under the same names as in AgentResult class.

        Args:
            input: The input for the agent run.
            options: The options for the agent run.
            context: The context for the agent run.
            tool_choice: Parameter that allows to control what tool is used at first call. Can be one of:
                - "auto": let model decide if tool call is needed
                - "none": do not call tool
                - "required: enforce tool usage (model decides which one)
                - Callable: one of provided tools

        Returns:
            A `StreamingResult` object for iteration and collection.

        Raises:
            AgentToolDuplicateError: If the tool names are duplicated.
            AgentToolNotSupportedError: If the selected tool type is not supported.
            AgentToolNotAvailableError: If the selected tool is not available.
            AgentInvalidPromptInputError: If the prompt/input combination is invalid.
            AgentMaxTurnsExceededError: If the maximum number of turns is exceeded.
        """
        generator = self._stream_internal(input, options, context, tool_choice)
        return AgentResultStreaming(generator)

    async def _stream_internal(
        self,
        input: str | PromptInputT | None = None,
        options: AgentOptions[LLMClientOptionsT] | None = None,
        context: AgentRunContext | None = None,
        tool_choice: ToolChoice | None = None,
    ) -> AsyncGenerator[str | ToolCall | ToolCallResult | SimpleNamespace | BasePrompt | Usage]:
        if context is None:
            context = AgentRunContext()

        input = cast(PromptInputT, input)
        merged_options = (self.default_options | options) if options else self.default_options
        llm_options = merged_options.llm_options or self.llm.default_options

        prompt_with_history = self._get_prompt_with_history(input)
        tools_mapping = await self._get_all_tools()
        turn_count = 0
        max_turns = merged_options.max_turns
        max_turns = 10 if max_turns is NOT_GIVEN else max_turns
        with trace(input=input, options=merged_options) as outputs:
            while not max_turns or turn_count < max_turns:
                returned_tool_call = False
                self._check_token_limits(merged_options, context.usage, prompt_with_history, self.llm)
                streaming_result = self.llm.generate_streaming(
                    prompt=prompt_with_history,
                    tools=[tool.to_function_schema() for tool in tools_mapping.values()],
                    tool_choice=tool_choice if tool_choice and turn_count == 0 else None,
                    options=self._get_llm_options(llm_options, merged_options, context.usage),
                )
                async for chunk in streaming_result:
                    yield chunk

                    if isinstance(chunk, ToolCall):
                        result = await self._execute_tool(tool_call=chunk, tools_mapping=tools_mapping, context=context)
                        yield result
                        prompt_with_history = prompt_with_history.add_tool_use_message(**result.__dict__)
                        returned_tool_call = True
                turn_count += 1
                if streaming_result.usage:
                    context.usage += streaming_result.usage

                if not returned_tool_call:
                    break
            else:
                raise AgentMaxTurnsExceededError(cast(int, max_turns))

            yield context.usage
            yield prompt_with_history
            if self.keep_history:
                self.history = prompt_with_history.chat
            yield outputs

    @staticmethod
    def _check_token_limits(
        options: AgentOptions[LLMClientOptionsT], usage: Usage, prompt: BasePrompt, llm: LLM[LLMClientOptionsT]
    ) -> None:
        if options.max_prompt_tokens or options.max_total_tokens:
            next_prompt_tokens = llm.count_tokens(prompt)
            if options.max_prompt_tokens and next_prompt_tokens > options.max_prompt_tokens - usage.prompt_tokens:
                raise AgentMaxTokensExceededError("prompt", options.max_prompt_tokens, next_prompt_tokens)
            if options.max_total_tokens and next_prompt_tokens > options.max_total_tokens - usage.total_tokens:
                raise AgentNextPromptOverLimitError(
                    "total", options.max_total_tokens, usage.total_tokens, next_prompt_tokens
                )

        if options.max_total_tokens and usage.total_tokens > options.max_total_tokens:
            raise AgentMaxTokensExceededError("total", options.max_total_tokens, usage.total_tokens)
        if options.max_prompt_tokens and usage.prompt_tokens > options.max_prompt_tokens:
            raise AgentMaxTokensExceededError("prompt", options.max_prompt_tokens, usage.prompt_tokens)
        if options.max_completion_tokens and usage.completion_tokens > options.max_completion_tokens:
            raise AgentMaxTokensExceededError("completion", options.max_completion_tokens, usage.completion_tokens)

    @staticmethod
    def _get_llm_options(
        llm_options: LLMClientOptionsT, options: AgentOptions[LLMClientOptionsT], usage: Usage
    ) -> LLMClientOptionsT:
        actual_limits: list[int] = [
            limit
            for limit in (options.max_total_tokens, options.max_prompt_tokens, options.max_completion_tokens)
            if isinstance(limit, int)
        ]

        if not actual_limits:
            return llm_options

        llm_options.max_tokens = min(actual_limits) - usage.total_tokens
        return llm_options

    def _prepare_synthesized_prompt(
        self, input: PromptInputT
    ) -> SimplePrompt | Prompt[PromptInputT, PromptOutputT] | None:
        curr_history = deepcopy(self.history)
        if (
            hasattr(self, "prompt_cls")
            and self.prompt_cls
            and isinstance(self.prompt_cls, type)
            and issubclass(self.prompt_cls, Prompt)
        ):
            prompt_cls = cast(type[Prompt[PromptInputT, PromptOutputT]], self.prompt_cls)
            if self.keep_history:
                self.prompt = cast(Prompt[PromptInputT, PromptOutputT], prompt_cls(input, curr_history))
                return self.prompt
            else:
                return cast(Prompt[PromptInputT, PromptOutputT], prompt_cls(input))

        if isinstance(self.prompt, type) and issubclass(self.prompt, Prompt):
            if self.keep_history:
                self.prompt = self.prompt(input, curr_history)
                return self.prompt
            else:
                return self.prompt(input)

        return None

    def _get_prompt_with_history(self, input: PromptInputT) -> SimplePrompt | Prompt[PromptInputT, PromptOutputT]:
        curr_history = deepcopy(self.history)
        new_prompt = self._prepare_synthesized_prompt(input)

        if new_prompt:
            return new_prompt

        if isinstance(self.prompt, Prompt):
            self.prompt.add_user_message(input)
            return self.prompt

        if isinstance(self.prompt, str) and isinstance(input, str):
            system_prompt = {"role": "system", "content": self.prompt}
            if len(curr_history) == 0:
                curr_history.append(system_prompt)
            else:
                system_idx = next((i for i, msg in enumerate(curr_history) if msg["role"] == "system"), -1)
                if system_idx == -1:
                    curr_history.insert(0, system_prompt)
                else:
                    curr_history[system_idx] = system_prompt
            incoming_user_prompt = input
        elif isinstance(self.prompt, str) and input is None:
            incoming_user_prompt = self.prompt
        elif isinstance(input, str):
            incoming_user_prompt = input
        else:
            raise AgentInvalidPromptInputError(self.prompt, input)

        curr_history.append({"role": "user", "content": incoming_user_prompt})

        return SimplePrompt(curr_history)

    async def _get_all_tools(self) -> dict[str, Tool]:
        tools_mapping = {}
        all_tools = list(self.tools)

        server_tools = await asyncio.gather(*[get_tools(server) for server in self.mcp_servers])
        for tools in server_tools:
            all_tools.extend(tools)

        for tool in all_tools:
            if tool.name in tools_mapping:
                raise AgentToolDuplicateError(tool.name)
            tools_mapping[tool.name] = tool

        return tools_mapping

    async def _execute_tool(
        self,
        tool_call: ToolCall,
        tools_mapping: dict[str, Tool],
        context: AgentRunContext | None = None,
    ) -> ToolCallResult:
        if tool_call.type != "function":
            raise AgentToolNotSupportedError(tool_call.type)

        if tool_call.name not in tools_mapping:
            raise AgentToolNotAvailableError(tool_call.name)

        tool = tools_mapping[tool_call.name]

        with trace(agent_id=self.id, tool_name=tool_call.name, tool_arguments=tool_call.arguments) as outputs:
            try:
                call_args = tool_call.arguments.copy()
                if tool.context_var_name:
                    call_args[tool.context_var_name] = context

                tool_output = (
                    await tool.on_tool_call(**call_args)
                    if iscoroutinefunction(tool.on_tool_call)
                    else tool.on_tool_call(**call_args)
                )

                if isinstance(tool_output, AgentResultStreaming):
                    async for _ in tool_output:
                        pass

                    tool_output = {
                        "content": tool_output.content,
                        "metadata": tool_output.metadata,
                        "tool_calls": tool_output.tool_calls,
                        "usage": tool_output.usage,
                    }

                outputs.result = {
                    "tool_output": tool_output,
                    "tool_call_id": tool_call.id,
                }

            except Exception as e:
                outputs.result = {
                    "error": str(e),
                    "tool_call_id": tool_call.id,
                }
                raise AgentToolExecutionError(tool_call.name, e) from e

        return ToolCallResult(
            id=tool_call.id,
            name=tool_call.name,
            arguments=tool_call.arguments,
            result=tool_output,
        )

    @requires_dependencies(["a2a.types"], "a2a")
    async def get_agent_card(
        self,
        name: str,
        description: str,
        version: str = "0.0.0",
        host: str = "127.0.0.1",
        port: int = 8000,
        protocol: str = "http",
        default_input_modes: list[str] | None = None,
        default_output_modes: list[str] | None = None,
        capabilities: "AgentCapabilities | None" = None,
        skills: list["AgentSkill"] | None = None,
    ) -> "AgentCard":
        """
        Create an AgentCard that encapsulates metadata about the agent,
        such as its name, version, description, network location, supported input/output modes,
        capabilities, and skills.

        Args:
            name: Human-readable name of the agent.
            description: A brief description of the agent.
            version: Version string of the agent. Defaults to "0.0.0".
            host: Hostname or IP where the agent will be served. Defaults to "0.0.0.0".
            port: Port number on which the agent listens. Defaults to 8000.
            protocol: URL scheme (e.g. "http" or "https"). Defaults to "http".
            default_input_modes: List of input content modes supported by the agent. Defaults to ["text"].
            default_output_modes: List of output content modes supported. Defaults to ["text"].
            capabilities: Agent capabilities; if None, defaults to empty capabilities.
            skills: List of AgentSkill objects representing the agent's skills.
                If None, attempts to extract skills from the agent's registered tools.

        Returns:
            An A2A-compliant agent descriptor including URL and capabilities.
        """
        return AgentCard(
            name=name,
            version=version,
            description=description,
            url=f"{protocol}://{host}:{port}",
            defaultInputModes=default_input_modes or ["text"],
            defaultOutputModes=default_output_modes or ["text"],
            skills=skills or await self._extract_agent_skills(),
            capabilities=capabilities or AgentCapabilities(),
        )

    async def _extract_agent_skills(self) -> list["AgentSkill"]:
        """
        The skill representation with name, id, description, and tags.
        """
        all_tools = await self._get_all_tools()
        return [
            AgentSkill(
                name=tool.name.replace("_", " ").title(),
                id=tool.name,
                description=f"{tool.description}\n\nParameters:\n{tool.parameters}",
                tags=[],
            )
            for tool in all_tools.values()
        ]

    @requires_dependencies("pydantic_ai")
    def to_pydantic_ai(self) -> "PydanticAIAgent":
        """
        Convert ragbits agent instance into a `pydantic_ai.Agent` representation.

        Returns:
            PydanticAIAgent: The equivalent Pydantic-based agent configuration.

        Raises:
            ValueError: If the `prompt` is not a string or a `Prompt` instance.
        """
        mcp_servers: list[mcp.MCPServerStdio | mcp.MCPServerHTTP] = []

        if not self.prompt:
            raise ValueError("Prompt is required but was None.")

        if isinstance(self.prompt, str):
            system_prompt = self.prompt
        else:
            if not self.prompt.system_prompt:
                raise ValueError("System prompt is required but was None.")
            system_prompt = self.prompt.system_prompt

        for mcp_server in self.mcp_servers:
            if isinstance(mcp_server, MCPServerStdio):
                mcp_servers.append(
                    mcp.MCPServerStdio(
                        command=mcp_server.params.command, args=mcp_server.params.args, env=mcp_server.params.env
                    )
                )
            elif isinstance(mcp_server, MCPServerStreamableHttp):
                timeout = mcp_server.params["timeout"]
                sse_timeout = mcp_server.params["sse_read_timeout"]

                mcp_servers.append(
                    mcp.MCPServerHTTP(
                        url=mcp_server.params["url"],
                        headers=mcp_server.params["headers"],
                        timeout=timeout.total_seconds() if isinstance(timeout, timedelta) else timeout,
                        sse_read_timeout=sse_timeout.total_seconds()
                        if isinstance(sse_timeout, timedelta)
                        else sse_timeout,
                    )
                )
        return PydanticAIAgent(
            model=self.llm.model_name,
            system_prompt=system_prompt,
            tools=[tool.to_pydantic_ai() for tool in self.tools],
            mcp_servers=mcp_servers,
        )

    @classmethod
    @requires_dependencies("pydantic_ai")
    def from_pydantic_ai(cls, pydantic_ai_agent: "PydanticAIAgent") -> Self:
        """
        Construct an agent instance from a `pydantic_ai.Agent` representation.

        Args:
            pydantic_ai_agent: A Pydantic-based agent configuration.

        Returns:
            An instance of the agent class initialized from the Pydantic representation.
        """
        mcp_servers: list[MCPServerStdio | MCPServerStreamableHttp] = []
        for mcp_server in pydantic_ai_agent._mcp_servers:
            if isinstance(mcp_server, mcp.MCPServerStdio):
                mcp_servers.append(
                    MCPServerStdio(
                        params={
                            "command": mcp_server.command,
                            "args": list(mcp_server.args),
                            "env": mcp_server.env or {},
                        }
                    )
                )
            elif isinstance(mcp_server, mcp.MCPServerHTTP):
                headers = mcp_server.headers or {}

                mcp_servers.append(
                    MCPServerStreamableHttp(
                        params={
                            "url": mcp_server.url,
                            "headers": {str(k): str(v) for k, v in headers.items()},
                            "sse_read_timeout": mcp_server.sse_read_timeout,
                            "timeout": mcp_server.timeout,
                        }
                    )
                )

        if not pydantic_ai_agent.model:
            raise ValueError("Missing LLM in `pydantic_ai.Agent` instance")
        elif isinstance(pydantic_ai_agent.model, str):
            model_name = pydantic_ai_agent.model
        else:
            model_name = pydantic_ai_agent.model.model_name

        return cls(
            llm=LiteLLM(model_name=model_name),  # type: ignore[arg-type]
            prompt="\n".join(pydantic_ai_agent._system_prompts),
            tools=[tool.function for _, tool in pydantic_ai_agent._function_tools.items()],
            mcp_servers=cast(list[MCPServer], mcp_servers),
        )

    def to_tool(self, name: str | None = None, description: str | None = None) -> Tool:
        """
        Convert the agent into a Tool instance.

        Args:
            name: Optional override for the tool name.
            description: Optional override for the tool description.

        Returns:
            Tool instance representing the agent.
        """
        return Tool.from_agent(self, name=name or self.name, description=description or self.description)

    @staticmethod
    def _make_prompt_class_for_agent_subclass(agent_cls: type["Agent[Any, Any, Any]"]) -> type[Prompt] | None:
        system = getattr(agent_cls, "system_prompt", None)
        if not system:
            raise ValueError(f"Agent {agent_cls.__name__} has no system_prompt")
        user = getattr(agent_cls, "user_prompt", "{{ input }}")

        input_type: type[BaseModel] | None = getattr(agent_cls, "input_type", None)
        if input_type is None:
            raise ValueError(f"Agent {agent_cls.__name__} cannot build a Prompt without input_type")

        base_generic = Prompt
        base = base_generic[input_type]

        attrs = {
            "system_prompt": system,
            "user_prompt": user,
        }

        return types.new_class(
            f"_{agent_cls.__name__}Prompt",
            (base,),
            exec_body=lambda ns: ns.update(attrs),
        )

    @overload
    @staticmethod
    def prompt_config(
        input_model: type[_Input],
    ) -> Callable[[type[Any]], type["Agent[LLMOptions, _Input, str]"]]: ...
    @overload
    @staticmethod
    def prompt_config(
        input_model: type[_Input],
        output_model: type[_Output],
    ) -> Callable[[type[Any]], type["Agent[LLMOptions, _Input, _Output]"]]: ...
    @staticmethod
    def prompt_config(
        input_model: type[_Input],
        output_model: type[_Output] | type[NotGiven] = NotGiven,
    ) -> Callable[[type[Any]], type["Agent[LLMOptions, _Input, _Output]"]]:
        """
        Decorator to bind both input and output types of an Agent subclass, with runtime checks.

        Raises:
            TypeError: if the decorated class is not a subclass of Agent,
                       or if input_model is not a Pydantic BaseModel.
        """
        if not isinstance(input_model, type) or not issubclass(input_model, BaseModel):
            raise TypeError(f"input_model must be a subclass of pydantic.BaseModel, got {input_model}")

        if not isinstance(output_model, type):
            raise TypeError(f"output_model must be a type, got {output_model}")

        def decorator(cls: type[Any]) -> type["Agent[LLMOptions, _Input, _Output]"]:
            if not isinstance(cls, type) or not issubclass(cls, Agent):
                raise TypeError(f"Can only decorate subclasses of Agent, got {cls}")

            cls.input_type = input_model
            cls.prompt_cls = Agent._make_prompt_class_for_agent_subclass(cls)

            return cast(type["Agent[LLMOptions, _Input, _Output]"], cls)

        return decorator



"""CLI commands for managing and running agents."""

import importlib.util
import sys
from pathlib import Path
from typing import Any

import typer

from ragbits.agents import Agent, ToolCallResult
from ragbits.agents._main import AgentResult

agents_app = typer.Typer(help="Commands for managing agents")


def import_agent_from_path(agent_path: str) -> Agent:
    """
    Import an agent from a module path.

    Args:
        agent_path: Path in format 'module.path:agent_variable' or 'path/to/file.py:agent_variable'

    Returns:
        The imported agent instance

    Raises:
        ValueError: If the path format is invalid
        ImportError: If the module cannot be imported
        AttributeError: If the agent variable is not found
    """
    if ":" not in agent_path:
        raise ValueError(
            f"Invalid agent path format: {agent_path}. "
            f"Expected 'module.path:agent_variable' or 'path/to/file.py:agent_variable'"
        )

    module_path, agent_name = agent_path.split(":", 1)

    # Check if it's a file path
    if module_path.endswith(".py") and (Path(module_path).exists() or "/" in module_path or "\\" in module_path):
        # Import from file
        file_path = Path(module_path).resolve()
        if not file_path.exists():
            raise ImportError(f"Agent file not found: {file_path}")

        spec = importlib.util.spec_from_file_location("agent_module", file_path)
        if spec is None or spec.loader is None:
            raise ImportError(f"Cannot load module from {file_path}")

        module = importlib.util.module_from_spec(spec)
        sys.modules["agent_module"] = module
        spec.loader.exec_module(module)
    else:
        # Import from module name
        try:
            module = importlib.import_module(module_path)
        except ImportError as e:
            raise ImportError(f"Cannot import module {module_path}: {e}") from e

    # Get the agent instance
    try:
        agent = getattr(module, agent_name)
    except AttributeError as e:
        raise AttributeError(f"Agent '{agent_name}' not found in module {module_path}") from e

    if not isinstance(agent, Agent):
        raise TypeError(f"'{agent_name}' is not an Agent instance, got {type(agent)}")

    return agent


def format_agent_metadata(agent: Agent) -> dict[str, Any]:
    """
    Extract and format agent metadata for display.

    Args:
        agent: The agent instance

    Returns:
        Dictionary containing formatted agent metadata
    """
    metadata = {
        "llm": {
            "model": getattr(agent.llm, "model_name", "Unknown"),
            "type": agent.llm.__class__.__name__,
        },
        "prompt": {
            "type": type(agent.prompt).__name__ if agent.prompt else "None",
            "has_system_prompt": hasattr(agent.prompt, "system_prompt") if agent.prompt else False,
        },
        "tools": {
            "count": len(agent.tools),
            "names": [tool.name for tool in agent.tools],
        },
        "mcp_servers": {
            "count": len(agent.mcp_servers),
            "types": [server.__class__.__name__ for server in agent.mcp_servers],
        },
        "history": {
            "enabled": agent.keep_history,
            "length": len(agent.history),
        },
    }

    return metadata


def run_interactive_agent(agent: Agent, agent_path: str) -> None:
    """
    Run the interactive TUI for the given agent.

    Args:
        agent: The agent instance to run interactively
        agent_path: The path used to import the agent (for display)
    """
    try:
        # Import textual components here to avoid dependency issues
        from textual import on
        from textual.app import App, ComposeResult
        from textual.containers import Horizontal, Vertical, VerticalScroll
        from textual.reactive import reactive
        from textual.widgets import (
            Button,
            Footer,
            Header,
            Input,
            LoadingIndicator,
            Log,
            Static,
            TabbedContent,
            TabPane,
            TextArea,
        )

        class AgentChatApp(App):
            """Interactive TUI application for chatting with ragbits agents."""

            CSS = """
            Screen {
                layout: vertical;
            }

            .chat-container {
                height: 1fr;
                border: solid $primary;
                margin: 1;
            }

            .input-container {
                height: auto;
                margin: 1;
            }

            .metadata-container {
                height: 1fr;
                border: solid $secondary;
                margin: 1;
            }

            .tools-log {
                height: 1fr;
                border: solid $accent;
            }

            .system-prompt {
                height: 1fr;
                border: solid $warning;
            }

            Input {
                margin: 1 0;
            }

            Button {
                margin: 0 1;
            }

            .controls {
                height: auto;
                margin: 1;
            }

            LoadingIndicator {
                height: 1;
                margin: 0 1;
            }
            """

            TITLE = "Ragbits Agent Interactive CLI"

            agent_path: reactive[str] = reactive("")
            current_conversation: reactive[list[dict[str, str]]] = reactive([])

            def __init__(self, agent: Agent, agent_path: str) -> None:
                super().__init__()
                self.agent = agent
                self.agent_path = agent_path
                self.metadata = format_agent_metadata(agent)

            def compose(self) -> ComposeResult:
                """Compose the app UI."""
                yield Header(show_clock=True)

                with TabbedContent(initial="chat"):
                    with TabPane("Chat", id="chat"):
                        yield Vertical(
                            Log(id="chat_log", classes="chat-container"),
                            LoadingIndicator(id="loading_indicator"),
                            Horizontal(
                                Input(placeholder="Type your message...", id="chat_input"),
                                Button("Send", variant="primary", id="send_button"),
                                classes="input-container",
                            ),
                            classes="chat-tab",
                        )

                    with TabPane("Agent Info", id="agent_info"):
                        yield VerticalScroll(
                            Static(self._format_agent_info(), id="agent_metadata"),
                            classes="metadata-container",
                        )

                    with TabPane("Tools Log", id="tools"):
                        yield Log(id="tools_log", classes="tools-log")

                    with TabPane("System Prompt", id="system_prompt"):
                        yield TextArea(
                            text=self._get_system_prompt(),
                            read_only=True,
                            id="system_prompt_area",
                            classes="system-prompt",
                        )

                with Horizontal(classes="controls"):
                    yield Button("Clear Chat", variant="warning", id="clear_button")
                    yield Button("Quit", variant="error", id="quit_button")

                yield Footer()

            def on_mount(self) -> None:
                """Handle app mount event."""
                chat_log = self.query_one("#chat_log", Log)
                chat_log.write_line(f"🤖 Agent loaded: {self.agent_path}")
                chat_log.write_line("💬 Type a message and press Enter or click Send to start chatting!")
                chat_log.write_line("💡 Tip: Type '/exit' to quit the interactive session")

                # Hide loading indicator initially
                loading = self.query_one("#loading_indicator", LoadingIndicator)
                loading.display = False

                # Focus on input
                self.query_one("#chat_input", Input).focus()

            @on(Input.Submitted, "#chat_input")
            @on(Button.Pressed, "#send_button")
            def handle_send_message(self, event: Input.Submitted | Button.Pressed) -> None:
                """Handle sending a message to the agent."""
                import asyncio

                chat_input = self.query_one("#chat_input", Input)
                message = chat_input.value.strip()

                if not message:
                    return

                # Clear input
                chat_input.value = ""

                # Check for /exit command
                if message.lower() == "/exit":
                    chat_log = self.query_one("#chat_log", Log)
                    chat_log.write_line("👋 Goodbye! Exiting...")
                    self.exit()
                    return

                # Add user message to chat
                chat_log = self.query_one("#chat_log", Log)
                chat_log.write_line(f"👤 You: {message}")

                # Run agent asynchronously
                asyncio.create_task(self._run_agent(message))

            @on(Button.Pressed, "#clear_button")
            def handle_clear_chat(self) -> None:
                """Clear the chat history."""
                chat_log = self.query_one("#chat_log", Log)
                chat_log.clear()
                chat_log.write_line("🤖 Chat cleared. Start a new conversation!")

                # Clear agent history if enabled
                if self.agent.keep_history:
                    self.agent.history = []

            @on(Button.Pressed, "#quit_button")
            def handle_quit(self) -> None:
                """Quit the application."""
                self.exit()

            async def _run_agent(self, message: str) -> None:
                """Run the agent with the provided message."""
                chat_log = self.query_one("#chat_log", Log)
                tools_log = self.query_one("#tools_log", Log)
                loading = self.query_one("#loading_indicator", LoadingIndicator)

                try:
                    # Show loading indicator while agent is thinking
                    loading.display = True
                    tools_log.write_line(f"🔧 Tool calls for message '{message[:50]}...':")

                    first_response = True
                    agent_results = self.agent.run_streaming(message)

                    async for result in agent_results:
                        match result:
                            case str():
                                if first_response:
                                    # Hide loading indicator and start agent response
                                    loading.display = False
                                    chat_log.write("\n🤖 Agent: ")
                                    first_response = False
                                chat_log.write(result)
                            case ToolCallResult():
                                tools_log.write_line(f"  • {result.name}({result.arguments}) → {result.result}")

                    total_usage = agent_results.usage
                    if first_response:
                        loading.display = False
                        chat_log.write("\n🤖 Agent: (No response)")

                    usage_text = (
                        f"📊 Usage: {total_usage.prompt_tokens}+"
                        f"{total_usage.completion_tokens}={total_usage.total_tokens} tokens"
                    )
                    chat_log.write_line(usage_text)

                except Exception as e:
                    # Always hide loading indicator on error
                    loading.display = False
                    chat_log.write_line(f"❌ Error: {e}")
                    tools_log.write_line(f"❌ Error running agent: {e}")

            def _format_agent_info(self) -> str:
                """Format agent metadata for display."""
                info = f"""🤖 Agent Information

        📍 Agent Path: {self.agent_path}

        🧠 LLM Configuration:
          • Model: {self.metadata['llm']['model']}
          • Type: {self.metadata['llm']['type']}

        📝 Prompt Configuration:
          • Type: {self.metadata['prompt']['type']}
          • Has System Prompt: {self.metadata['prompt']['has_system_prompt']}

        🔧 Tools:
          • Count: {self.metadata['tools']['count']}
          • Names: {', '.join(self.metadata['tools']['names']) if self.metadata['tools']['names'] else 'None'}

        🌐 MCP Servers:
          • Count: {self.metadata['mcp_servers']['count']}
          • Types: {', '.join(self.metadata['mcp_servers']['types'])
                     if self.metadata['mcp_servers']['types'] else 'None'}

        💾 History:
          • Enabled: {self.metadata['history']['enabled']}
          • Current Length: {self.metadata['history']['length']}
        """
                return info

            def _get_system_prompt(self) -> str:
                """Get the system prompt text."""
                if not self.agent.prompt:
                    return "No system prompt configured"

                if isinstance(self.agent.prompt, str):
                    return self.agent.prompt

                if hasattr(self.agent.prompt, "system_prompt"):
                    return self.agent.prompt.system_prompt or "No system prompt found in prompt class"

                return f"Prompt type: {type(self.agent.prompt).__name__} (cannot extract system prompt text)"

        app = AgentChatApp(agent, agent_path)
        app.run()

    except ImportError as e:
        typer.echo(
            "❌ Error: textual is required for interactive mode. Install with 'pip install ragbits-agents[cli]'",
            err=True,
        )
        typer.echo(f"Details: {e}")
        raise typer.Exit(1) from e


@agents_app.command()
def run(
    agent_path: str = typer.Argument(
        help="Path to agent in format 'module.path:agent_variable' (e.g., 'examples.city_explorer:agent')"
    ),
    interactive: bool = typer.Option(
        True, "--interactive/--no-interactive", "-i", help="Run in interactive mode with TUI"
    ),
) -> None:
    """
    Run an agent interactively or in batch mode.

    AGENT_PATH should be in format 'module.path:agent_variable'
    Example: 'examples.city_explorer:city_explorer_agent'
    """
    try:
        agent = import_agent_from_path(agent_path)

        if interactive:
            run_interactive_agent(agent, agent_path)
        else:
            typer.echo(f"Running agent {agent_path} in batch mode (not implemented yet)")

    except Exception as e:
        typer.echo(f"Error running agent: {e}", err=True)
        raise typer.Exit(1) from e


@agents_app.command(name="exec")
def execute_agent(
    agent_path: str = typer.Argument(help="Path to agent in format 'module.path:agent_variable'"),
    input_text: str = typer.Argument(help="Input text to send to the agent"),
) -> None:
    """
    Execute a single input against an agent and return the result.

    This runs the agent once with the provided input and outputs the result.
    """
    import asyncio

    try:
        agent = import_agent_from_path(agent_path)

        typer.echo(f"🤖 Running agent: {agent_path}")
        typer.echo(f"💬 Input: {input_text}")
        typer.echo("⏳ Processing...")

        # Run the agent
        async def run_single() -> AgentResult[Any]:
            agent_result = await agent.run(input_text)
            return agent_result

        result = asyncio.run(run_single())

        typer.echo("\n✅ Response:")
        typer.echo(result.content)

        if result.tool_calls:
            typer.echo(f"\n🔧 Tools used: {len(result.tool_calls)}")
            for tool_call in result.tool_calls:
                typer.echo(f"  • {tool_call.name}: {tool_call.result}")

        if result.usage.total_tokens > 0:
            usage_text = (
                f"\n📊 Token usage: {result.usage.prompt_tokens}+"
                f"{result.usage.completion_tokens}={result.usage.total_tokens}"
            )
            typer.echo(usage_text)

    except Exception as e:
        typer.echo(f"❌ Error executing agent: {e}", err=True)
        raise typer.Exit(1) from e


def register(app: typer.Typer) -> None:
    """
    Register the CLI commands for the agents package.

    Args:
        app: The Typer object to register the commands with.
    """
    app.add_typer(agents_app, name="agents", help="Commands for managing and running agents")



from urllib.parse import urlparse

import uvicorn
from a2a.types import AgentCard
from fastapi import FastAPI
from pydantic import BaseModel

from ragbits.agents import Agent, AgentResult


class Request(BaseModel):
    """
    Represents a request to the agent server.
    """

    params: dict


class Response(BaseModel):
    """
    Represents a response from the agent server.
    """

    result: dict | None = None
    error: str | None = None


def _format_response(result: AgentResult) -> dict:
    return {
        "content": result.content,
        "metadata": result.metadata,
        "history": result.history,
        "tool_calls": [t.__dict__ for t in result.tool_calls or []],
    }


def create_agent_app(agent: Agent, agent_card: "AgentCard", input_model: type[BaseModel]) -> "FastAPI":
    """
    Create a FastAPI server for the given agent and input schema.
    The server exposes:
        - GET /.well-known/agent.json: Returns the agent's metadata.
        - POST /: Accepts structured input and returns the agent's output.

    Args:
        agent: An instance of the ragbits Agent to expose.
        agent_card: Metadata describing the agent.
        input_model: A Pydantic model describing the input format.

    Returns:
        FastAPI: The configured FastAPI app.
    """
    app = FastAPI(title=agent_card.name, description=agent_card.description)

    @app.get("/.well-known/agent.json", response_model=AgentCard)
    async def get_card() -> AgentCard:
        return agent_card

    @app.post("/", response_model=Response)
    async def handle(request: Request) -> Response:
        try:
            input = input_model(**request.params)
            result = await agent.run(input)

            return Response(result=_format_response(result))

        except Exception as e:
            return Response(error=str(e))

    return app


def create_agent_server(
    agent: Agent,
    agent_card: "AgentCard",
    input_model: type[BaseModel],
) -> "uvicorn.Server":
    """
    Create a Uvicorn server instance that serves the specified agent over HTTP.

    The server's host and port are extracted from the URL in the given agent_card.

    Args:
        agent: The Ragbits Agent instance to serve.
        agent_card: Metadata for the agent, including its URL.
        input_model: A Pydantic model class used to validate incoming request data.

    Returns:
        A configured uvicorn.Server instance ready to be started.

    Raises:
        ValueError: If the URL in agent_card does not contain a valid hostname or port.
    """
    app = create_agent_app(agent=agent, agent_card=agent_card, input_model=input_model)
    url = urlparse(agent_card.url)

    if not url.hostname:
        raise ValueError(f"Could not parse hostname from URL: {agent_card.url}")
    if not url.port:
        raise ValueError(f"Could not parse port from URL: {agent_card.url}")

    config = uvicorn.Config(app=app, host=url.hostname, port=url.port)
    server = uvicorn.Server(config=config)

    return server



from contextlib import suppress
from functools import partial
from typing import Any

from ragbits.agents.mcp.server import MCPServer
from ragbits.agents.tool import Tool

with suppress(ImportError):
    from mcp import Tool as MCPTool
    from mcp.types import CallToolResult


async def get_tools(server: MCPServer) -> list[Tool]:
    """
    Get all function tools from a single MCP server.

    Args:
        server: The MCP server to retrieve tools from.

    Returns:
        List of Tool instances from the server.
    """
    tools = await server.list_tools()
    return [
        Tool(
            name=tool.name,
            description=tool.description,
            # MCP spec doesn't require the inputSchema to have `properties`, but OpenAI spec does.
            parameters={**tool.inputSchema, "properties": tool.inputSchema.get("properties", {})},
            on_tool_call=partial(call_mcp_tool, server, tool),
        )
        for tool in tools
    ]


async def call_mcp_tool(server: MCPServer, tool: "MCPTool", **arguments: Any) -> "CallToolResult":  # noqa: ANN401
    """
    Call an MCP tool.

    Args:
        server: The MCP server containing the tool.
        tool: The MCP tool to call.
        **arguments: Keyword arguments to pass to the tool.

    Returns:
        The result of the tool execution.
    """
    return await server.call_tool(tool.name, arguments)



from ragbits.agents.tools.openai import get_code_interpreter_tool, get_image_generation_tool, get_web_search_tool

__all__ = ["get_code_interpreter_tool", "get_image_generation_tool", "get_web_search_tool"]



from typing import Any, Literal


class AgentError(Exception):
    """
    Base class for all exceptions raised by the Agent.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class AgentToolNotSupportedError(AgentError):
    """
    Raised when the selected tool type is not supported.
    """

    def __init__(self, tool_type: str) -> None:
        super().__init__(f"The tool call type in LLM response is not supported: {tool_type}")
        self.tool_type = tool_type


class AgentToolNotAvailableError(AgentError):
    """
    Raised when the selected tool is not available.
    """

    def __init__(self, tool_name: str) -> None:
        super().__init__(f"Selected tool is not available: {tool_name}")
        self.tool_name = tool_name


class AgentToolExecutionError(AgentError):
    """
    Raised when the tool execution fails.
    """

    def __init__(self, tool_name: str, error: Exception) -> None:
        super().__init__(f"Tool execution failed: {tool_name}, error: {error}")
        self.tool_name = tool_name
        self.error = error


class AgentToolDuplicateError(AgentError):
    """
    Raised when agent tool names are duplicated.
    """

    def __init__(self, tool_name: str) -> None:
        super().__init__(f"Duplicate tool name found: {tool_name}")
        self.tool_name = tool_name


class AgentInvalidPromptInputError(AgentError):
    """
    Raised when the prompt/input combination is invalid.
    """

    def __init__(self, prompt: Any, input: Any) -> None:  # noqa: ANN401
        super().__init__(f"Invalid prompt/input combination: prompt={prompt}, input={input}")
        self.prompt_type = prompt
        self.input_type = input


class AgentMaxTurnsExceededError(AgentError):
    """
    Raised when the maximum number of turns is exceeded.
    """

    def __init__(self, max_turns: int) -> None:
        super().__init__(
            f"The number of Agent turns exceeded the limit of {max_turns}."
            "To change this limit, pass ragbits.agents.AgentOptions with max_turns when initializing the Agent."
            "agent = Agent(options=AgentOptions(max_turns=x))"
        )
        self.max_turns = max_turns


class AgentMaxTokensExceededError(AgentError):
    """
    Raised when the maximum number of total tokens is exceeded.
    """

    def __init__(self, limit_type: Literal["total", "prompt", "completion"], limit: int, actual: int) -> None:
        super().__init__(f"The number of {limit_type} tokens exceeded the limit of {limit}, actual: {actual}.")
        self.limit_type = limit_type
        self.limit = limit
        self.actual = actual


class AgentNextPromptOverLimitError(AgentError):
    """
    Raised when the next prompt won't fit under the limit.
    """

    def __init__(
        self, limit_type: Literal["total", "prompt"], limit: int, actual: int, next_prompt_tokens: int
    ) -> None:
        super().__init__(
            f"The next prompt won't fit under the limit of {limit} {limit_type} tokens, "
            f"actual: {actual}, next_prompt_tokens: {next_prompt_tokens}.",
        )

        self.limit_type = limit_type
        self.limit = limit
        self.actual = actual
        self.next_prompt_tokens = next_prompt_tokens



from mcp.types import Tool as MCPTool
from pydantic import BaseModel

from ragbits.agents.mcp.utils import call_mcp_tool, get_tools
from ragbits.agents.tool import Tool

from .helpers import FakeMCPServer


class Foo(BaseModel):
    bar: str
    baz: int


class Bar(BaseModel):
    qux: dict[str, str]


async def test_get_all_function_tools():
    names = ["test_tool_1", "test_tool_2", "test_tool_3"]
    schemas = [
        {},
        Foo.model_json_schema(),
        Bar.model_json_schema(),
    ]

    server1 = FakeMCPServer()
    server1.add_tool(names[0], schemas[0])
    server1.add_tool(names[1], schemas[1])
    server1.add_tool(names[2], schemas[2])

    tools = await get_tools(server1)
    assert len(tools) == 3
    assert all(tool.name in names for tool in tools)

    for idx, tool in enumerate(tools):
        assert isinstance(tool, Tool)
        if schemas[idx] == {}:
            assert tool.parameters == {"properties": {}}
        else:
            assert tool.parameters == schemas[idx]
        assert tool.name == names[idx]


async def test_call_mcp_tool():
    server = FakeMCPServer()
    server.add_tool("test_tool_1", {})

    tool = MCPTool(name="test_tool_1", inputSchema={})

    await call_mcp_tool(server, tool)
    # Just making sure it doesn't crash


async def test_util_adds_properties():
    schema = {
        "type": "object",
        "description": "Test tool",
    }

    server = FakeMCPServer()
    server.add_tool("test_tool", schema)

    tools = await get_tools(server)
    tool = next(tool for tool in tools if tool.name == "test_tool")

    assert isinstance(tool, Tool)
    assert "properties" in tool.parameters
    assert tool.parameters["properties"] == {}

    assert tool.parameters == {"type": "object", "description": "Test tool", "properties": {}}



from typing import Any, TypeVar

from pydantic import BaseModel

from ragbits.agents._main import Agent
from ragbits.core.llms.base import LLMClientOptionsT

QuestionAnswerPromptInputT = TypeVar("QuestionAnswerPromptInputT", bound="QuestionAnswerPromptInput")
QuestionAnswerPromptOutputT = TypeVar("QuestionAnswerPromptOutputT", bound="QuestionAnswerPromptOutput | str")

QuestionAnswerAgent = Agent[LLMClientOptionsT, QuestionAnswerPromptInputT, QuestionAnswerPromptOutputT]


class QuestionAnswerPromptInput(BaseModel):
    """
    Input for the question answer agent.
    """

    question: str
    """The question to answer."""
    context: Any | None = None
    """The context to answer the question."""


class QuestionAnswerPromptOutput(BaseModel):
    """
    Output for the question answer agent.
    """

    answer: str
    """The answer to the question."""



from ragbits.agents.mcp.server import (
    MCPServer,
    MCPServerSse,
    MCPServerSseParams,
    MCPServerStdio,
    MCPServerStdioParams,
    MCPServerStreamableHttp,
    MCPServerStreamableHttpParams,
)

__all__ = [
    "MCPServer",
    "MCPServerSse",
    "MCPServerSseParams",
    "MCPServerStdio",
    "MCPServerStdioParams",
    "MCPServerStreamableHttp",
    "MCPServerStreamableHttpParams",
]



from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

from pydantic import BaseModel
from typing_extensions import Self

from ragbits.core.llms.base import LLMClientOptionsT
from ragbits.core.prompt.prompt import PromptInputT, PromptOutputT
from ragbits.core.utils.decorators import requires_dependencies
from ragbits.core.utils.function_schema import convert_function_to_function_schema, get_context_variable_name

if TYPE_CHECKING:
    from ragbits.agents import Agent, AgentResultStreaming

with suppress(ImportError):
    from pydantic_ai import Tool as PydanticAITool


@dataclass
class ToolCallResult:
    """
    Result of the tool call.
    """

    id: str
    """Unique identifier for the specific tool call instance."""
    name: str
    """Name of the tool that was called."""
    arguments: dict[str, Any]
    """Dictionary containing the arguments passed to the tool"""
    result: Any
    """The output from the tool call."""


@dataclass
class Tool:
    """
    Function tool that can be called by the agent.
    """

    name: str
    """The name of the tool/function."""
    description: str | None
    """Optional description of what the tool does."""
    parameters: dict[str, Any]
    """Dictionary containing the parameters JSON schema."""
    on_tool_call: Callable
    """The actual callable function to execute when the tool is called."""
    context_var_name: str | None = None
    """The name of the context variable that this tool accepts."""

    @classmethod
    def from_callable(cls, callable: Callable) -> Self:
        """
        Create a Tool instance from a callable function.

        Args:
            callable: The function to convert into a Tool

        Returns:
            A new Tool instance representing the callable function.
        """
        schema = convert_function_to_function_schema(callable)

        return cls(
            name=schema["function"]["name"],
            description=schema["function"]["description"],
            parameters=schema["function"]["parameters"],
            on_tool_call=callable,
            context_var_name=get_context_variable_name(callable),
        )

    def to_function_schema(self) -> dict[str, Any]:
        """
        Convert the Tool to a standardized function schema format.

        Returns:
            Function schema dictionary with 'type' and 'function' keys.
        """
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters,
            },
        }

    @requires_dependencies("pydantic_ai")
    def to_pydantic_ai(self) -> "PydanticAITool":
        """
        Convert ragbits tool to a Pydantic AI Tool.

        Returns:
            A `pydantic_ai.tools.Tool` object.
        """
        return PydanticAITool(
            function=self.on_tool_call,
            name=self.name,
            description=self.description,
        )

    @classmethod
    def from_agent(
        cls,
        agent: "Agent[LLMClientOptionsT, PromptInputT, PromptOutputT]",
        name: str | None = None,
        description: str | None = None,
    ) -> "Tool":
        """
        Wraps a downstream agent as a single tool. The tool parameters are inferred from
        the downstream agent's prompt input.

        Args:
            agent: The downstream agent to wrap as a tool.
            name: Optional override for the tool name.
            description: Optional override for the tool description.

        Returns:
            Tool instance representing the agent.
        """
        display_name = name or agent.name or "agent"
        variable_name = display_name.replace(" ", "_").lower()
        description = description or agent.description

        input_model_cls = getattr(agent.prompt, "input_type", None)
        if input_model_cls and issubclass(input_model_cls, BaseModel):
            fields = input_model_cls.model_fields
            properties = {}
            required = list(fields.keys())

            for field_name in fields:
                param_desc = None
                for t in getattr(agent, "tools", []):
                    t_params = getattr(t, "parameters", {}).get("properties", {})
                    if field_name in t_params:
                        param_desc = t_params[field_name].get("description")
                        break

                properties[field_name] = {
                    "type": "string",
                    "title": field_name.capitalize(),
                    "description": param_desc,
                }
        else:
            properties = {"input": {"type": "string", "description": "Input for the downstream agent"}}
            required = ["input"]

        parameters = {"type": "object", "properties": properties, "required": required}

        def _on_tool_call(**kwargs: dict) -> "AgentResultStreaming":
            if input_model_cls and issubclass(input_model_cls, BaseModel):
                model_input = input_model_cls(**kwargs)
            else:
                model_input = kwargs.get("input")

            return agent.run_streaming(model_input)

        return cls(
            name=variable_name,
            description=description,
            parameters=parameters,
            on_tool_call=_on_tool_call,
            context_var_name=get_context_variable_name(agent.run),
        )


ToolChoice = Literal["auto", "none", "required"] | Callable



import json
from collections.abc import Callable
from typing import cast

import pytest
from pydantic import BaseModel

from ragbits.agents import Agent, AgentRunContext
from ragbits.agents._main import AgentOptions, AgentResult, AgentResultStreaming, ToolCallResult, ToolChoice
from ragbits.agents.exceptions import (
    AgentInvalidPromptInputError,
    AgentMaxTurnsExceededError,
    AgentToolNotAvailableError,
    AgentToolNotSupportedError,
)
from ragbits.core.llms.base import Usage, UsageItem
from ragbits.core.llms.mock import MockLLM, MockLLMOptions
from ragbits.core.prompt.prompt import Prompt


class WeatherInput(BaseModel):
    location: str
    date: str


class WeatherPrompt(Prompt[WeatherInput, str]):
    system_prompt = "You are a weather assistant"
    user_prompt = "What's the weather like in {{location}} on {{date}}?"


class CustomPrompt(Prompt):
    user_prompt = "Custom test prompt"


def get_weather(location: str) -> str:
    """
    Returns the current weather for a given location.

    Args:
        location: The location to get the weather for.

    Returns:
        The current weather for the given location.
    """
    if "san francisco" in location.lower():
        return json.dumps({"location": "San Francisco", "temperature": "72", "unit": "fahrenheit"})
    else:
        return json.dumps({"location": location, "temperature": "unknown"})


def get_weather_context(location: str, context: AgentRunContext | None) -> AgentRunContext | None:  # noqa: D417
    """
    Returns the current weather for a given location.

    Args:
        location: The location to get the weather for.

    Returns:
        The current weather for the given location.
    """
    return context


@pytest.fixture
def llm_without_tool_call() -> MockLLM:
    options = MockLLMOptions(response="Test LLM output")
    return MockLLM(default_options=options)


@pytest.fixture
def llm_with_tool_call() -> MockLLM:
    options = MockLLMOptions(
        response="Temperature is 72 fahrenheit",
        tool_calls=[
            {
                "name": "get_weather",
                "arguments": '{"location": "San Francisco"}',
                "id": "test",
                "type": "function",
            }
        ],
    )
    return MockLLM(default_options=options)


@pytest.fixture
def llm_wrong_tool_type() -> MockLLM:
    options = MockLLMOptions(
        response="Temperature is 72 fahrenheit",
        tool_calls=[
            {
                "name": "get_weather",
                "arguments": '{"location": "San Francisco"}',
                "id": "test",
                "type": "tool",
            }
        ],
    )
    return MockLLM(default_options=options)


@pytest.fixture
def llm_multiple_tool_calls() -> MockLLM:
    options = MockLLMOptions(
        response="Final response after multiple tool calls",
        tool_calls=[
            {
                "name": "get_weather",
                "arguments": '{"location": "San Francisco"}',
                "id": "test1",
                "type": "function",
            },
            {
                "name": "get_weather",
                "arguments": '{"location": "New York"}',
                "id": "test2",
                "type": "function",
            },
            {
                "name": "get_weather",
                "arguments": '{"location": "London"}',
                "id": "test3",
                "type": "function",
            },
        ],
    )
    return MockLLM(default_options=options)


@pytest.fixture
def llm_with_tool_call_context() -> MockLLM:
    options = MockLLMOptions(
        response="Temperature is 72 fahrenheit",
        tool_calls=[
            {
                "name": "get_weather_context",
                "arguments": '{"location": "San Francisco"}',
                "id": "test",
                "type": "function",
            }
        ],
    )
    return MockLLM(default_options=options)


def get_time() -> str:
    """
    Returns the current time.

    Returns:
        The current time as a string.
    """
    return "12:00 PM"


@pytest.fixture
def llm_no_tool_call_when_none() -> MockLLM:
    """LLM that doesn't call tools when tool_choice is 'none'."""
    options = MockLLMOptions(response="I cannot call tools right now.")
    return MockLLM(default_options=options)


@pytest.fixture
def llm_auto_tool_call() -> MockLLM:
    """LLM that automatically decides to call a tool."""
    options = MockLLMOptions(
        response="Let me check the weather for you.",
        tool_calls=[
            {
                "name": "get_weather",
                "arguments": '{"location": "New York"}',
                "id": "auto_test",
                "type": "function",
            }
        ],
    )
    return MockLLM(default_options=options)


@pytest.fixture
def llm_required_tool_call() -> MockLLM:
    """LLM that is forced to call a tool when tool_choice is 'required'."""
    options = MockLLMOptions(
        response="",
        tool_calls=[
            {
                "name": "get_weather",
                "arguments": '{"location": "Boston"}',
                "id": "required_test",
                "type": "function",
            }
        ],
    )
    return MockLLM(default_options=options)


@pytest.fixture
def llm_specific_tool_call() -> MockLLM:
    """LLM that calls a specific tool when tool_choice is a specific function."""
    options = MockLLMOptions(
        response="",
        tool_calls=[
            {
                "name": "get_time",
                "arguments": "{}",
                "id": "specific_test",
                "type": "function",
            }
        ],
    )
    return MockLLM(default_options=options)


async def _run(
    agent: Agent,
    input: str | BaseModel | None = None,
    options: AgentOptions | None = None,
    context: AgentRunContext | None = None,
    tool_choice: ToolChoice | None = None,
) -> AgentResult:
    return await agent.run(input, options=options, context=context, tool_choice=tool_choice)


async def _run_streaming(
    agent: Agent,
    input: str | BaseModel | None = None,
    options: AgentOptions | None = None,
    context: AgentRunContext | None = None,
    tool_choice: ToolChoice | None = None,
) -> AgentResultStreaming:
    result = agent.run_streaming(input, options=options, context=context, tool_choice=tool_choice)
    async for _chunk in result:
        pass
    return result


@pytest.mark.parametrize(("method", "result_type"), [(_run, AgentResult), (_run_streaming, AgentResultStreaming)])
async def test_agent_run_no_tools(llm_without_tool_call: MockLLM, method: Callable, result_type: type):
    """Test a simple run of the agent without tools."""
    agent = Agent(
        llm=llm_without_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent)

    assert isinstance(result, result_type)
    result = cast(AgentResult, result)
    assert result.content == "Test LLM output"
    assert result.tool_calls is None


@pytest.mark.parametrize(("method", "result_type"), [(_run, AgentResult), (_run_streaming, AgentResultStreaming)])
async def test_agent_run_tools(llm_with_tool_call: MockLLM, method: Callable, result_type: type):
    """Test a simple run of the agent without tools."""
    agent = Agent(
        llm=llm_with_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent)

    assert isinstance(result, result_type)
    result = cast(AgentResult, result)
    assert result.content == "Temperature is 72 fahrenheit"
    assert result.tool_calls == [
        ToolCallResult(
            id="test",
            name="get_weather",
            arguments={
                "location": "San Francisco",
            },
            result='{"location": "San Francisco", "temperature": "72", "unit": "fahrenheit"}',
        ),
    ]


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_raises_when_wrong_tool_returned(llm_with_tool_call: MockLLM, method: Callable):
    def fake_func() -> None: ...

    agent = Agent(
        llm=llm_with_tool_call,
        prompt=CustomPrompt,
        tools=[fake_func],
    )

    with pytest.raises(AgentToolNotAvailableError):
        await method(agent)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_raises_when_wrong_tool_type(llm_wrong_tool_type: MockLLM, method: Callable):
    agent = Agent(
        llm=llm_wrong_tool_type,
        prompt=CustomPrompt,
        tools=[get_weather],
    )

    with pytest.raises(AgentToolNotSupportedError):
        await method(agent)


@pytest.mark.parametrize(
    ("method", "context"),
    [(_run, None), (_run, AgentRunContext()), (_run_streaming, None), (_run_streaming, AgentRunContext())],
)
async def test_agent_run_tools_with_context(
    llm_with_tool_call_context: MockLLM, method: Callable, context: AgentRunContext | None
):
    agent = Agent(
        llm=llm_with_tool_call_context,
        prompt=CustomPrompt,
        tools=[get_weather_context],
    )

    result = await method(agent, context=context)

    assert result.content == "Temperature is 72 fahrenheit"
    assert result.tool_calls[0].id == "test"
    assert result.tool_calls[0].name == "get_weather_context"
    assert result.tool_calls[0].arguments == {"location": "San Francisco"}
    assert isinstance(result.tool_calls[0].result, AgentRunContext)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_context_is_updated(llm_without_tool_call: MockLLM, method: Callable):
    context: AgentRunContext = AgentRunContext()
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="NOT IMPORTANT",
    )
    _ = await method(agent, context=context)
    assert context.usage == Usage(
        requests=[
            UsageItem(
                model="mock:mock",
                prompt_tokens=10,
                completion_tokens=20,
                total_tokens=30,
                estimated_cost=0.0,
            )
        ]
    )


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_raises_invalid_prompt_input_error_with_none_prompt_and_none_input(
    llm_without_tool_call: MockLLM, method: Callable
):
    agent = Agent(llm=llm_without_tool_call, prompt=None)  # type: ignore

    with pytest.raises(AgentInvalidPromptInputError) as exc_info:
        await method(agent)

    assert exc_info.value.prompt_type is None
    assert exc_info.value.input_type is None
    assert "Invalid prompt/input combination: prompt=None, input=None" in str(exc_info.value)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_raises_invalid_prompt_input_error_with_invalid_types(llm_without_tool_call: MockLLM, method: Callable):
    agent = Agent(llm=llm_without_tool_call, prompt=123)  # type: ignore

    with pytest.raises(AgentInvalidPromptInputError) as exc_info:
        await method(agent, input={"key": "value"})

    assert exc_info.value.prompt_type == 123
    assert exc_info.value.input_type == {"key": "value"}
    assert "Invalid prompt/input combination" in str(exc_info.value)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_raises_invalid_prompt_input_error_with_list_input(llm_without_tool_call: MockLLM, method: Callable):
    agent = Agent(llm=llm_without_tool_call, prompt=None)  # type: ignore

    with pytest.raises(AgentInvalidPromptInputError) as exc_info:
        await method(agent, input=["item1", "item2"])

    assert exc_info.value.prompt_type is None
    assert exc_info.value.input_type == ["item1", "item2"]
    assert "Invalid prompt/input combination" in str(exc_info.value)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_with_initial_history(llm_without_tool_call: MockLLM, method: Callable):
    """Test agent with initial history."""
    initial_history = [
        {"role": "system", "content": "You are a helpful assistant"},
        {"role": "user", "content": "Hello"},
        {"role": "assistant", "content": "Hi there!"},
    ]

    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="You are a weather assistant",
        history=initial_history,
        keep_history=True,
    )

    await method(agent, input="What's the weather like?")

    assert agent.history

    assert len(agent.history) == 5  # system + initial user + assistant + new user + new assistant
    assert agent.history[0]["role"] == "system"
    assert agent.history[0]["content"] == "You are a weather assistant"  #! Updated system prompt
    assert agent.history[1]["role"] == "user"
    assert agent.history[1]["content"] == "Hello"
    assert agent.history[2]["role"] == "assistant"
    assert agent.history[2]["content"] == "Hi there!"
    assert agent.history[3]["role"] == "user"
    assert agent.history[3]["content"] == "What's the weather like?"
    assert agent.history[4]["role"] == "assistant"
    assert agent.history[4]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_without_keep_history(llm_without_tool_call: MockLLM, method: Callable):
    """Test agent without history preservation."""
    initial_history = [
        {"role": "user", "content": "Previous message"},
        {"role": "assistant", "content": "Previous response"},
    ]

    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="System prompt",
        history=initial_history,
        keep_history=False,
    )

    result = await method(agent, input="New message")

    # Agent history should not be updated
    assert agent.history == initial_history
    # But result history should be updated
    assert len(result.history) == 5  # system + user + assistant + new user + new assistant
    assert result.history[0]["role"] == "system"
    assert result.history[0]["content"] == "System prompt"
    assert result.history[1]["role"] == "user"
    assert result.history[1]["content"] == "Previous message"
    assert result.history[2]["role"] == "assistant"
    assert result.history[2]["content"] == "Previous response"
    assert result.history[3]["role"] == "user"
    assert result.history[3]["content"] == "New message"
    assert result.history[4]["role"] == "assistant"
    assert result.history[4]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_string_prompt_and_input(llm_without_tool_call: MockLLM, method: Callable):
    """Test history handling with string prompt and string input."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="You are a helpful assistant",
        keep_history=True,
    )

    await method(agent, input="Hello")

    assert agent.history is not None
    assert len(agent.history) == 3  # system + user + assistant
    assert agent.history[0]["role"] == "system"
    assert agent.history[0]["content"] == "You are a helpful assistant"
    assert agent.history[1]["role"] == "user"
    assert agent.history[1]["content"] == "Hello"
    assert agent.history[2]["role"] == "assistant"
    assert agent.history[2]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_string_prompt_no_input(llm_without_tool_call: MockLLM, method: Callable):
    """Test history handling with string prompt and no input."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="Tell me a joke",
        keep_history=True,
    )

    await method(agent)

    assert agent.history is not None
    assert len(agent.history) == 2  # user (prompt) + assistant
    assert agent.history[0]["role"] == "user"
    assert agent.history[0]["content"] == "Tell me a joke"
    assert agent.history[1]["role"] == "assistant"
    assert agent.history[1]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_weather_prompt_with_history_multiple_runs(
    llm_without_tool_call: MockLLM, method: Callable
):
    """Test history handling with custom prompt class with input type."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt=WeatherPrompt,
        keep_history=True,
    )

    input_data = WeatherInput(location="London", date="today")
    await method(agent, input=input_data)

    input_data2 = WeatherInput(location="Paris", date="tomorrow")
    await method(agent, input=input_data2)

    assert agent.history is not None
    assert len(agent.history) == 5
    assert agent.history[0]["role"] == "system"
    assert agent.history[0]["content"] == "You are a weather assistant"
    assert agent.history[1]["role"] == "user"
    assert agent.history[1]["content"] == "What's the weather like in London on today?"
    assert agent.history[2]["role"] == "assistant"
    assert agent.history[2]["content"] == "Test LLM output"
    assert agent.history[3]["role"] == "user"
    assert agent.history[3]["content"] == "What's the weather like in Paris on tomorrow?"
    assert agent.history[4]["role"] == "assistant"
    assert agent.history[4]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_weather_prompt_no_history_multiple_runs(
    llm_without_tool_call: MockLLM, method: Callable
):
    """Test history handling with custom prompt class with input type."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt=WeatherPrompt,
        keep_history=False,
    )

    input_data = WeatherInput(location="London", date="today")
    result = await method(agent, input=input_data)
    assert agent.history == []
    assert result.history[0]["role"] == "system"
    assert result.history[0]["content"] == "You are a weather assistant"
    assert result.history[1]["role"] == "user"
    assert result.history[1]["content"] == "What's the weather like in London on today?"
    assert result.history[2]["role"] == "assistant"
    assert result.history[2]["content"] == "Test LLM output"

    input_data2 = WeatherInput(location="Paris", date="tomorrow")
    result2 = await method(agent, input=input_data2)
    assert agent.history == []
    assert result2.history[0]["role"] == "system"
    assert result2.history[0]["content"] == "You are a weather assistant"
    assert result2.history[1]["role"] == "user"
    assert result2.history[1]["content"] == "What's the weather like in Paris on tomorrow?"
    assert result2.history[2]["role"] == "assistant"
    assert result2.history[2]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_no_prompt_string_input(llm_without_tool_call: MockLLM, method: Callable):
    """Test history handling with no prompt and string input."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt=None,
        keep_history=True,
    )

    await method(agent, input="Direct message")

    assert agent.history is not None
    assert len(agent.history) == 2  # user + assistant
    assert agent.history[0]["role"] == "user"
    assert agent.history[0]["content"] == "Direct message"
    assert agent.history[1]["role"] == "assistant"
    assert agent.history[1]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_string_prompt_multiple_runs(llm_without_tool_call: MockLLM, method: Callable):
    """Test history accumulation across multiple runs."""
    agent: Agent = Agent(
        llm=llm_without_tool_call,
        prompt="You are a helpful assistant",
        keep_history=True,
    )

    result1 = await method(agent, input="First message")
    assert result1.history is not None
    assert len(result1.history) == 3  # system + user + assistant

    result2 = await method(agent, input="Second message")
    assert result2.history is not None
    assert len(result2.history) == 5  # system + user1 + assistant1 + user2 + assistant2

    # Verify the conversation flow
    assert result2.history[0]["role"] == "system"
    assert result2.history[1]["role"] == "user"
    assert result2.history[1]["content"] == "First message"
    assert result2.history[2]["role"] == "assistant"
    assert result2.history[2]["content"] == "Test LLM output"
    assert result2.history[3]["role"] == "user"
    assert result2.history[3]["content"] == "Second message"
    assert result2.history[4]["role"] == "assistant"
    assert result2.history[4]["content"] == "Test LLM output"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_history_with_tools(llm_with_tool_call: MockLLM, method: Callable):
    """Test history handling with tool calls."""
    agent: Agent = Agent(
        llm=llm_with_tool_call,
        prompt="You are a weather assistant",
        tools=[get_weather],
        keep_history=True,
    )

    await method(agent, input="What's the weather in San Francisco?")

    assert len(agent.history) == 5

    assert agent.history[0]["role"] == "system"
    assert agent.history[1]["role"] == "user"
    assert agent.history[2]["role"] == "assistant"
    assert agent.history[2]["content"] is None  # Tool call message
    assert "tool_calls" in agent.history[2]
    assert agent.history[3]["role"] == "tool"
    assert agent.history[4]["role"] == "assistant"
    assert agent.history[4]["content"] == "Temperature is 72 fahrenheit"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_max_turns_exceeded(llm_with_tool_call: MockLLM, method: Callable):
    """Test that AgentMaxToolCallsExceededError is raised when max_tool_calls is exceeded in a single response."""
    agent = Agent(
        llm=llm_with_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
        default_options=AgentOptions(max_turns=1),
    )

    options: AgentOptions = AgentOptions(llm_options=None)

    with pytest.raises(AgentMaxTurnsExceededError) as exc_info:
        await method(agent, options=options)

    assert exc_info.value.max_turns == 1
    assert "The number of Agent turns exceeded the limit of 1" in str(exc_info.value)


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_max_turns_not_exeeded_with_many_tool_calls(llm_multiple_tool_calls: MockLLM, method: Callable):
    """Test that AgentMaxToolCallsExceededError is raised when max_tool_calls is exceeded in a single response."""
    agent = Agent(
        llm=llm_multiple_tool_calls,
        prompt=CustomPrompt,
        tools=[get_weather],
        default_options=AgentOptions(max_turns=2),
    )

    options: AgentOptions = AgentOptions(llm_options=None)

    result = await method(agent, options=options)

    assert result.content == "Final response after multiple tool calls"
    assert len(result.tool_calls) == 3


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_with_tool_choice_none(llm_no_tool_call_when_none: MockLLM, method: Callable):
    """Test agent run with tool_choice set to 'none'."""
    agent = Agent(
        llm=llm_no_tool_call_when_none,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent, tool_choice="none")

    assert result.content == "I cannot call tools right now."
    assert result.tool_calls is None


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_with_auto_tool_call(llm_auto_tool_call: MockLLM, method: Callable):
    """Test agent run with automatic tool call."""
    agent = Agent(
        llm=llm_auto_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent)

    assert result.content == "Let me check the weather for you."
    assert len(result.tool_calls) == 1
    assert result.tool_calls[0].id == "auto_test"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_with_required_tool_call(llm_required_tool_call: MockLLM, method: Callable):
    """Test agent run with required tool call."""
    agent = Agent(
        llm=llm_required_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent, tool_choice="required")

    assert result.content == ""
    assert len(result.tool_calls) == 1
    assert result.tool_calls[0].id == "required_test"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_with_specific_tool_call(llm_specific_tool_call: MockLLM, method: Callable):
    """Test agent run with specific tool call."""
    agent = Agent(
        llm=llm_specific_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather, get_time],
    )
    result = await method(agent, tool_choice=get_time)

    assert result.content == ""
    assert len(result.tool_calls) == 1
    assert result.tool_calls[0].id == "specific_test"
    assert result.tool_calls[0].name == "get_time"
    assert result.tool_calls[0].result == "12:00 PM"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_agent_run_with_tool_choice_auto_explicit(llm_auto_tool_call: MockLLM, method: Callable):
    """Test agent run with tool_choice explicitly set to 'auto'."""
    agent = Agent(
        llm=llm_auto_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather],
    )
    result = await method(agent, tool_choice="auto")

    assert result.content == "Let me check the weather for you."
    assert len(result.tool_calls) == 1
    assert result.tool_calls[0].name == "get_weather"
    assert result.tool_calls[0].arguments == {"location": "New York"}


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_tool_choice_with_multiple_tools_available(llm_auto_tool_call: MockLLM, method: Callable):
    """Test tool_choice behavior when multiple tools are available."""
    agent = Agent(
        llm=llm_auto_tool_call,
        prompt=CustomPrompt,
        tools=[get_weather, get_time],  # Multiple tools available
    )

    result = await method(agent, tool_choice="auto")

    assert result.content == "Let me check the weather for you."
    assert len(result.tool_calls) == 1
    # The LLM chose to call get_weather based on its configuration
    assert result.tool_calls[0].name == "get_weather"


@pytest.mark.parametrize("method", [_run, _run_streaming])
async def test_tool_choice_history_preservation(llm_with_tool_call: MockLLM, method: Callable):
    """Test that tool_choice works correctly with history preservation."""
    agent: Agent = Agent(
        llm=llm_with_tool_call,
        prompt="You are a helpful assistant",
        tools=[get_weather],
        keep_history=True,
    )

    await method(agent, input="Check weather", tool_choice="auto")
    assert len(agent.history) >= 3  # At least system, user, assistant messages
    # Should include tool call in history
    tool_call_messages = [msg for msg in agent.history if msg.get("role") == "tool"]
    assert len(tool_call_messages) >= 1


async def test_explicit_input_type_prompt_creation():
    class CustomInput(BaseModel):
        foo: int
        bar: str

    @Agent.prompt_config(CustomInput)
    class AgentWithExplicitInput(Agent):
        system_prompt = "System prompt"
        user_prompt = "{{ foo }} {{ bar }}"

    prompt_cls = AgentWithExplicitInput.prompt_cls
    assert prompt_cls is not None
    assert issubclass(prompt_cls, Prompt)
    assert prompt_cls.system_prompt == "System prompt"
    assert prompt_cls.user_prompt == "{{ foo }} {{ bar }}"


async def test_default_user_prompt_is_input_placeholder():
    class CustomInputModel(BaseModel):
        input: int

    @Agent.prompt_config(CustomInputModel)
    class AgentExplicitPrompt(Agent):
        system_prompt = "Explicit system"

    prompt_cls2 = AgentExplicitPrompt.prompt_cls
    assert prompt_cls2 is not None
    assert issubclass(prompt_cls2, Prompt)
    assert prompt_cls2.user_prompt == "{{ input }}"


async def test_input_type_check_with_system_prompt(llm_with_tool_call: MockLLM):
    class AgentExplicitPrompt(Agent):
        system_prompt = "Explicit system"

    with pytest.raises(ValueError):
        AgentExplicitPrompt(
            llm=llm_with_tool_call,
            prompt="You are a helpful assistant",
            tools=[get_weather],
            keep_history=True,
        )



from ragbits.agents._main import (
    Agent,
    AgentDependencies,
    AgentOptions,
    AgentResult,
    AgentResultStreaming,
    AgentRunContext,
    ToolCallResult,
)
from ragbits.agents.types import QuestionAnswerAgent, QuestionAnswerPromptInput, QuestionAnswerPromptOutput

__all__ = [
    "Agent",
    "AgentDependencies",
    "AgentOptions",
    "AgentResult",
    "AgentResultStreaming",
    "AgentRunContext",
    "QuestionAnswerAgent",
    "QuestionAnswerPromptInput",
    "QuestionAnswerPromptOutput",
    "ToolCallResult",
]



import base64
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from ragbits.agents.tools.openai import OpenAITools
from ragbits.agents.tools.types import ImageGenerationResponse


@pytest.fixture
def mock_responses_llm() -> MagicMock:
    """Fixture for mocking OpenAIResponsesLLM."""
    mock = MagicMock()
    mock.use_tool = AsyncMock()
    return mock


@patch("ragbits.agents.tools.openai.AsyncOpenAI")
async def test_search_web(mock_async_openai: MagicMock, mock_responses_llm: MagicMock) -> None:
    """Test that search_web calls use_tool and returns output_text."""
    mock_response = MagicMock()
    mock_response.output_text = "Test output"
    mock_responses_llm.use_tool.return_value = mock_response
    tools = OpenAITools("test_model", None, tool_param=MagicMock())
    tools._responses_llm = mock_responses_llm

    result = await tools.search_web("test query")

    mock_responses_llm.use_tool.assert_called_once_with("test query")
    assert result == mock_response


@patch("ragbits.agents.tools.openai.AsyncOpenAI")
async def test_code_interpreter(mock_async_openai: MagicMock, mock_responses_llm: MagicMock) -> None:
    """Test that code_interpreter calls use_tool and returns output_text."""
    mock_response = MagicMock()
    mock_response.output_text = "Interpreter output"
    mock_responses_llm.use_tool.return_value = mock_response
    tools = OpenAITools("test_model", None, tool_param=MagicMock())
    tools._responses_llm = mock_responses_llm

    result = await tools.code_interpreter("run code")

    mock_responses_llm.use_tool.assert_called_once_with("run code")
    assert result == mock_response


@patch("ragbits.agents.tools.openai.AsyncOpenAI")
@patch("builtins.open", new_callable=MagicMock)
async def test_image_generation(
    mock_open: MagicMock, mock_async_openai: MagicMock, mock_responses_llm: MagicMock
) -> None:
    """Test image_generation saves image and returns correct text."""
    image_data = base64.b64encode(b"test_image_content").decode("utf-8")

    mock_output_item = MagicMock()
    mock_output_item.type = "image_generation_call"
    mock_output_item.result = image_data

    mock_response = MagicMock()
    mock_response.output_text = "Generated image."
    mock_response.output = [mock_output_item]

    mock_file_handle = MagicMock()
    mock_open.return_value.__enter__.return_value = mock_file_handle

    mock_responses_llm.use_tool.return_value = mock_response
    tools = OpenAITools("test_model", None, tool_param=MagicMock())
    tools._responses_llm = mock_responses_llm

    result = await tools.image_generation("a cat", save_path="cat.png")

    mock_responses_llm.use_tool.assert_called_once_with("a cat")
    mock_open.assert_called_once_with("cat.png", "wb")
    mock_file_handle.write.assert_called_once_with(b"test_image_content")

    expected_result = ImageGenerationResponse(image_path="cat.png", output_text="Generated image.")
    assert result == expected_result


@patch("ragbits.agents.tools.openai.AsyncOpenAI")
async def test_image_generation_no_image(mock_async_openai: MagicMock, mock_responses_llm: MagicMock) -> None:
    """Test image_generation when no image is returned."""
    mock_output_item = MagicMock()
    mock_output_item.type = "image_generation_call"
    mock_output_item.result = None

    mock_response = MagicMock()
    mock_response.output_text = "No image was generated."
    mock_response.output = [mock_output_item]

    mock_responses_llm.use_tool.return_value = mock_response
    tools = OpenAITools("test_model", None, tool_param=MagicMock())
    tools._responses_llm = mock_responses_llm

    result = await tools.image_generation("a dog")
    expected_result = ImageGenerationResponse(image_path=None, output_text="No image was generated.")
    assert result == expected_result



from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field


class ImageGenerationResponse(BaseModel):
    """Result from image generation tool."""

    model_config = ConfigDict(extra="allow")  # Allow extra fields

    image_path: Path | str | None = Field(description="Path to the generated image file, None if generation failed")
    output_text: str = Field(description="Text output from the image generation process")
